//+------------------------------------------------------------------+
//|                                                    incMatrix.mq4 |
//|                                                                * |
//|                                                                * |
//+------------------------------------------------------------------+
#property copyright "Integer"
#property link      "https://login.mql5.com/ru/users/Integer"
//+------------------------------------------------------------------+
//|                              |
//+------------------------------------------------------------------+
/*

       , :   ,   .. 
         :     .
   
   :
   
   double m[]={1,2,3,
               4,5,6,
               2,3}; //       .
*/

class CIntMatrix
  {

public:
   //+------------------------------------------------------------------+
   //|      aA                                |
   //|   aRows -  , aCols -             |
   //+------------------------------------------------------------------+
   void SetSize(double  &aA[],int aRows,int aCols)
     {
      int               m_sz=aRows*aCols+2;
      ArrayResize(aA,m_sz);
      ArrayInitialize(aA,0);
      aA[m_sz-2]=aRows;
      aA[m_sz-1]=aCols;
     }
   //+------------------------------------------------------------------+
   //|     (Value)   aA,            |
   //|      aRow   aCol                    |
   //+------------------------------------------------------------------+
   void SetValue(double  &aA[],int aRow,int aCol,double aValue)
     {
      aA[aRow*GetCols(aA)+aCol]=aValue;
     }
   //+------------------------------------------------------------------+
   //|       aA.                    |
   //|     :                                          |
   //|   aRows -  , aCols -             |
   //+------------------------------------------------------------------+
   int GetSize(double  &aA[],int  &aRows,int  &aCols)
     {
      int               m_sz=ArraySize(aA);
      m_sz-=2;
      aRows=(int)aA[m_sz];
      aCols=(int)aA[m_sz+1];
      return(m_sz);
     }
   //+------------------------------------------------------------------+
   //|       aA                         |
   //+------------------------------------------------------------------+
   int GetRows(double  &aA[])
     {
      return((int)aA[ArraySize(aA)-2]);
     }
   //+------------------------------------------------------------------+
   //|       aA                       |
   //+------------------------------------------------------------------+
   int GetCols(double  &aA[])
     {
      return((int)aA[ArraySize(aA)-1]);
     }
   //+------------------------------------------------------------------+
   //|       aA,           |
   //|     aRow   aCol                                   |
   //+------------------------------------------------------------------+
   double GetValue(double  &aA[],int aRow,int aCol)
     {
      return(aA[aRow*GetCols(aA)+aCol]);
     }
   //+------------------------------------------------------------------+
   //|       aFrom   aTo                  |
   //+------------------------------------------------------------------+
   void Copy(double  &aFrom[],double  &aTo[])
     {
      ArrayResize(aTo,ArraySize(aFrom));
      ArrayCopy(aTo,aFrom,0,0,ArraySize(aFrom));
     }
   //+------------------------------------------------------------------+
   //|                | 
   //|   (     )                    |
   //+------------------------------------------------------------------+
   bool CheckForAdd(double  &aA[],double  &aB[])
     {
      int               m_ARows,m_ACols,m_BRows,m_BCols;
      GetSize(aA,m_ARows,m_ACols);
      GetSize(aB,m_BRows,m_BCols);
      return(m_ARows==m_BRows && m_ACols==m_BCols);
     }
   //+------------------------------------------------------------------+
   //|               | 
   //|   (     )                    |
   //+------------------------------------------------------------------+
   bool CheckForMult(double  &aA[],double  &aB[])
     {
      return(GetCols(aA)==GetRows(aB));
     }
   //+------------------------------------------------------------------+
   //|   ,                           |
   //+------------------------------------------------------------------+
   bool CheckIsSq(double  &aA[])
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      return(m_ARows==m_ACols);
     }
   //+------------------------------------------------------------------+
   //|     aNum   aA.                            |
   //|    ()     c aR.       |
   //+------------------------------------------------------------------+
   void AddNum(double  &aA[],double aNum,double  &aR[])
     {
      int               m_sz=ArraySize(aA);
      ArrayResize(aR,m_sz);
      aR[m_sz-2]=aA[m_sz-2];
      aR[m_sz-1]=aA[m_sz-1];
      for(int m_i=ArraySize(aA)-3;m_i>=0;m_i--)
        {
         aR[m_i]=aA[m_i]+aNum;
        }
     }
   //+------------------------------------------------------------------+
   //|     aA   aNum.                             | 
   //|    ()     c aR.       |
   //+------------------------------------------------------------------+
   void MultNum(double  &aA[],double aNum,double  &aR[])
     {
      int               m_sz=ArraySize(aA);
      ArrayResize(aR,m_sz);
      aR[m_sz-2]=aA[m_sz-2];
      aR[m_sz-1]=aA[m_sz-1];
      for(int m_i=ArraySize(aA)-3;m_i>=0;m_i--)
        {
         aR[m_i]=aA[m_i]*aNum;
        }
     }
   //+------------------------------------------------------------------+
   //|     aA  aB.                                    |
   //|    ()     c aAB.      |
   //+------------------------------------------------------------------+
   void AddMx(double  &aA[],double  &aB[],double  &aAB[])
     {
      int               m_ARows,m_ACols,m_BRows,m_BCols,m_ABRows,m_ABCols;
      GetSize(aA,m_ARows,m_ACols);
      GetSize(aB,m_BRows,m_BCols);
      m_ABRows=MathMin(m_ARows,m_BRows);
      m_ABCols=MathMin(m_ACols,m_BCols);
      double            m_tmp[];
      SetSize(m_tmp,m_ABRows,m_ABCols);
      for(int m_r=0;m_r<m_ABRows;m_r++)
        {
         for(int m_c=0;m_c<m_ABCols;m_c++)
           {
            int               m_i=m_r*m_ABCols+m_c;
            m_tmp[m_i]=aA[m_i]+aB[m_i];
           }
        }
      Copy(m_tmp,aAB);
     }
   //+------------------------------------------------------------------+
   //|     aA   aB.                             |
   //|    ()     c aAB.      |
   //+------------------------------------------------------------------+
   void MultMx(double  &aA[],double  &aB[],double  &aAB[])
     {
      int               m_ARows,m_ACols,m_BRows,m_BCols,m_ABRows,m_ABCols;
      GetSize(aA,m_ARows,m_ACols);
      GetSize(aB,m_BRows,m_BCols);
      m_ABRows=m_ARows;
      m_ABCols=m_BCols;
      int               m_N=MathMin(m_ACols,m_BRows);
      double            m_tmp[];
      SetSize(m_tmp,m_ABRows,m_ABCols);
      for(int m_r=0;m_r<m_ABRows;m_r++)
        {
         for(int m_c=0;m_c<m_ABCols;m_c++)
           {
            double            m_sum=0;
            for(int m_n=0;m_n<m_N;m_n++)
              {
               m_sum+=GetValue(aA,m_r,m_n)*GetValue(aB,m_n,m_c);
              }
            SetValue(m_tmp,m_r,m_c,m_sum);
           }
        }
      Copy(m_tmp,aAB);
     }
   //+------------------------------------------------------------------+
   //|     aA.              |
   //|       c aT                            |
   //+------------------------------------------------------------------+
   void Transpose(double  &aA[],double  &aT[])
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      double            m_tmp[];
      SetSize(m_tmp,m_ACols,m_ARows);
      for(int m_r=0;m_r<m_ARows;m_r++)
        {
         for(int m_c=0;m_c<m_ACols;m_c++)
           {
            double            m_Val=GetValue(aA,m_r,m_c);
            SetValue(m_tmp,m_c,m_r,m_Val);
           }
        }
      Copy(m_tmp,aT);
     }
   //+------------------------------------------------------------------+
   //|      .                   |
   //|   aA -  ,                                         |
   //|   aAA -   (  )       |
   //+------------------------------------------------------------------+
   void AlgAdd(double  &aA[],double  &aAA[])
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      double            m_tmp[];
      Copy(aA,m_tmp);
      for(int m_r=0;m_r<m_ARows;m_r++)
        {
         for(int m_c=0;m_c<m_ACols;m_c++)
           {
            if((m_r+m_c)%2==1)
              {
               double            m_Val=GetValue(m_tmp,m_r,m_c);
               m_Val*=-1;
               SetValue(m_tmp,m_r,m_c,m_Val);
              }
           }
        }
      Copy(m_tmp,aAA);
     }
   //+------------------------------------------------------------------+
   //|      aR   aA  .        |
   //|      true       |
   //|    false,                      |
   //+------------------------------------------------------------------+
   bool Invert(double  &aA[],double  &aR[])
     {
      double            m_D=Def(aA);
      if(m_D==0)return(false);
      m_D=1.0/m_D;
      double            m_tmp[];
      MinorDefMx(aA,m_tmp);
      AlgAdd(m_tmp,m_tmp);
      Transpose(m_tmp,m_tmp);
      MultNum(m_tmp,m_D,m_tmp);
      Copy(m_tmp,aR);
      return(true);
     }
   //+------------------------------------------------------------------+
   //|      aT   aA        |
   //+------------------------------------------------------------------+
   void Triangle(double  &aA[],double  &aT[])
     {
      double            m_A[];
      Copy(aA,m_A);
      int               m_Rows,m_Cols;
      GetSize(m_A,m_Rows,m_Cols);
      for(int m_r=0;m_r<m_Rows;m_r++)
        {
         double            m_vrr=GetValue(m_A,m_r,m_r);
         if(m_vrr==0)
           {
            for(int m_r2=m_r+1;m_r2<m_Rows;m_r2++)
              {
               if(GetValue(m_A,m_r2,m_r)!=0)
                 {
                  for(int m_c=0;m_c<m_Cols;m_c++)
                    {
                     double            m_v=GetValue(m_A,m_r,m_c);
                     SetValue(m_A,m_r,m_c,GetValue(m_A,m_r2,m_c));
                     SetValue(m_A,m_r2,m_c,m_v);
                    }
                 }
              }
            m_vrr=GetValue(m_A,m_r,m_r);
           }
         if(m_vrr!=0)
           {
            for(int m_r2=m_r+1;m_r2<m_Rows;m_r2++)
              {
               double            m_vr0=GetValue(m_A,m_r2,m_r);
               double            m_d=(-1)*m_vr0/m_vrr;
               for(int m_c=0;m_c<m_Cols;m_c++)
                 {
                  double            m_v0=GetValue(m_A,m_r,m_c);
                  double            m_vr=GetValue(m_A,m_r2,m_c);
                  m_vr=m_vr+m_v0*m_d;
                  SetValue(m_A,m_r2,m_c,m_vr);
                 }
              }
           }
        }
      Copy(m_A,aT);
     }
   //+------------------------------------------------------------------+
   //|       aA   aRow   aCol.    |
   //|         aM                      |
   //+------------------------------------------------------------------+
   void Minor(double  &aA[],int aRow,int aCol,double  &aM[])
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      int               m_MRows=m_ARows-1;
      int               m_MCols=m_ACols-1;
      double            m_tmp[];
      SetSize(m_tmp,m_MRows,m_MCols);
      int               m_r,m_c;
      double            m_Val;
      for(m_r=0;m_r<aRow;m_r++)
        {
         for(m_c=0;m_c<aCol;m_c++)
           {
            m_Val=GetValue(aA,m_r,m_c);
            SetValue(m_tmp,m_r,m_c,m_Val);
           }
         for(m_c=aCol;m_c<m_MCols;m_c++)
           {
            m_Val=GetValue(aA,m_r,m_c+1);
            SetValue(m_tmp,m_r,m_c,m_Val);
           }
        }
      for(m_r=aRow;m_r<m_MRows;m_r++)
        {
         for(m_c=0;m_c<aCol;m_c++)
           {
            m_Val=GetValue(aA,m_r+1,m_c);
            SetValue(m_tmp,m_r,m_c,m_Val);
           }
         for(m_c=aCol;m_c<m_MCols;m_c++)
           {
            m_Val=GetValue(aA,m_r+1,m_c+1);
            SetValue(m_tmp,m_r,m_c,m_Val);
           }
        }
      Copy(m_tmp,aM);
     }
   //+------------------------------------------------------------------+
   //|        aA             |
   //|     aRow   aCol                                  |
   //+------------------------------------------------------------------+
   double MinorDef(double  &aA[],int aRow,int aCol)
     {
      double            m_tmp[];
      Minor(aA,aRow,aCol,m_tmp);
      return(Def(m_tmp));
     }
   //+------------------------------------------------------------------+
   //|      (                  |
   //|    ).                                        |
   //|   aA -  ,                                         |
   //|   aM -     (  ) |
   //+------------------------------------------------------------------+
   void MinorDefMx(double  &aA[],double  &aM[])
     {
      double            m_tmp[];
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      SetSize(m_tmp,m_ARows,m_ACols);
      for(int m_r=0;m_r<m_ARows;m_r++)
        {
         for(int m_c=0;m_c<m_ACols;m_c++)
           {
            double            m_mdef=MinorDef(aA,m_r,m_c);
            SetValue(m_tmp,m_r,m_c,m_mdef);
           }
        }
      Copy(m_tmp,aM);
     }
   //+------------------------------------------------------------------+
   //|       aA                    |
   //+------------------------------------------------------------------+
   double Def(double  &aA[])
     {
      double            m_A[];
      Copy(aA,m_A);
      int               m_Rows,m_Cols;
      GetSize(m_A,m_Rows,m_Cols);
      int               m_MSize=MathMin(m_Rows,m_Cols);
      double            m_Sn[];
      ArrayResize(m_Sn,m_MSize);
      ArrayInitialize(m_Sn,1);
      for(int m_s=0;m_s<m_MSize;m_s++)
        {
         double            m_v00=GetValue(m_A,m_s,m_s);
         if(m_v00==0)
           {
            for(int m_r=m_s+1;m_r<m_MSize;m_r++)
              {
               double            m_vr0=GetValue(m_A,m_r,m_s);
               if(m_vr0!=0)
                 {
                  for(int m_c=m_s;m_c<m_MSize;m_c++)
                    {
                     double            m_v1=GetValue(m_A,m_s,m_c);
                     double            m_v2=GetValue(m_A,m_r,m_c);
                     SetValue(m_A,m_s,m_c,m_v2);
                     SetValue(m_A,m_r,m_c,m_v1);
                    }
                  m_Sn[m_s]=-1;
                  break;
                 }
              }
            m_v00=GetValue(m_A,m_s,m_s);
            if(m_v00==0)
              {
               return(0);
              }
           }
         for(int m_r=m_s+1;m_r<m_MSize;m_r++)
           {
            double            m_vr=GetValue(m_A,m_r,m_s);
            if(m_vr==0)continue;
            double            m_d=(-1)*m_vr/m_v00;
            for(int m_c=0;m_c<m_MSize;m_c++)
              {
               double            m_v0=GetValue(m_A,m_s,m_c);
               m_vr=GetValue(m_A,m_r,m_c);
               m_vr=m_vr+m_v0*m_d;
               SetValue(m_A,m_r,m_c,m_vr);
              }
           }
        }
      double            m_def=1;
      for(int m_s=0;m_s<m_MSize;m_s++)
        {
         m_def*=m_Sn[m_s]*GetValue(m_A,m_s,m_s);
        }
      return(m_def);
     }
   //+------------------------------------------------------------------+
   //|      aA                                     |
   //+------------------------------------------------------------------+
   int Rank(double  &aA[])
     {
      double            m_d;
      int               m_r,m_c;
      return(RankDRC(aA,m_d,m_r,m_c));
     }
   //+------------------------------------------------------------------+
   //|      aA    :            |
   //|   aDef -  ,                                  |
   //|   aRow -    ,   0              |
   //|   aCol -    ,   0             |
   //+------------------------------------------------------------------+
   int RankDRC(double  &aA[],double  &aDef,int  &aRow,int  &aCol)
     {
      int               m_Rows,m_Cols;
      GetSize(aA,m_Rows,m_Cols);
      int               m_MSize=MathMin(m_Rows,m_Cols);
      double            m_A[];
      for(int m_s=m_MSize;m_s>0;m_s--)
        {
         SetSize(m_A,m_s,m_s);
         for(int m_r=0;m_r<=m_Rows-m_s;m_r++)
           {
            for(int m_c=0;m_c<=m_Cols-m_s;m_c++)
              {
               for(int m_r2=0;m_r2<m_s;m_r2++)
                 {
                  for(int m_c2=0;m_c2<m_s;m_c2++)
                    {
                     double            m_v=GetValue(aA,m_r+m_r2,m_c+m_c2);
                     SetValue(m_A,m_r2,m_c2,m_v);
                    }
                 }
               aDef=Def(m_A);
               if(aDef!=0)
                 {
                  aRow=m_r;
                  aCol=m_c;
                  return(m_s);
                 }
              }
           }
        }
      return(0);
     }
   //+------------------------------------------------------------------+
   //|       aFromCol   aFrom          |
   //|     aTo     aToCol.                     |
   //|         aR                   |
   //+------------------------------------------------------------------+
   void CopyCol(double  &aFrom[],double  &aTo[],int aFromCol,int aToCol,double  &aR[])
     {
      int               m_FromRows=GetRows(aFrom);
      int               m_ToRows=GetRows(aTo);
      int               m_Rows=MathMin(m_FromRows,m_ToRows);
      double            m_tmp[];
      Copy(aTo,m_tmp);
      for(int m_r=0;m_r<m_Rows;m_r++)
        {
         double            m_Val=GetValue(aFrom,m_r,aFromCol);
         SetValue(m_tmp,m_r,aToCol,m_Val);
        }
      Copy(m_tmp,aR);
     }
   //+------------------------------------------------------------------+
   //|       aFromRow   aFrom           |
   //|     aTo     aToRow.                      |
   //|         aR                   |
   //+------------------------------------------------------------------+
   void CopyRow(double  &aFrom[],double  &aTo[],int aFromRow,int aToRow,double  &aR[])
     {
      int               m_FromCols=GetCols(aFrom);
      int               m_ToCols=GetCols(aTo);
      int               m_Cols=MathMin(m_FromCols,m_ToCols);
      double            m_tmp[];
      Copy(aTo,m_tmp);
      for(int m_c=0;m_c<m_Cols;m_c++)
        {
         double            m_Val=GetValue(aFrom,aFromRow,m_c);
         SetValue(m_tmp,aToRow,m_c,m_Val);
        }
      Copy(m_tmp,aR);
     }
   //+------------------------------------------------------------------+
   //|     aA      aC.        |
   //|        c aF                  |
   //+------------------------------------------------------------------+
   void AppendCol(double  &aA[],double  &aC[],double  &aF[])
     {
      int               m_Rows,m_Cols;
      GetSize(aA,m_Rows,m_Cols);
      double            m_tmp[];
      SetSize(m_tmp,m_Rows,m_Cols+1);
      for(int m_r=0;m_r<m_Rows;m_r++)
        {
         for(int m_c=0;m_c<m_Cols;m_c++)
           {
            SetValue(m_tmp,m_r,m_c,GetValue(aA,m_r,m_c));
           }
        }
      m_Rows=MathMin(m_Rows,GetRows(aC));
      for(int m_r=0;m_r<m_Rows;m_r++)
        {
         SetValue(m_tmp,m_r,m_Cols,GetValue(aC,m_r,0));
        }
      Copy(m_tmp,aF);
     }
   //+------------------------------------------------------------------+
   //|     aA      aR.         |
   //|        c aF                  |
   //+------------------------------------------------------------------+
   void AppendRow(double  &aA[],double  &aR[],double  &aF[])
     {
      int               m_Rows,m_Cols;
      GetSize(aA,m_Rows,m_Cols);
      double            m_tmp[];
      SetSize(m_tmp,m_Rows+1,m_Cols);
      for(int m_r=0;m_r<m_Rows;m_r++)
        {
         for(int m_c=0;m_c<m_Cols;m_c++)
           {
            SetValue(m_tmp,m_r,m_c,GetValue(aA,m_r,m_c));
           }
        }
      m_Cols=MathMin(m_Cols,GetCols(aR));
      for(int m_c=0;m_c<m_Cols;m_c++)
        {
         SetValue(m_tmp,m_Rows,m_c,GetValue(aR,0,m_c));
        }
      Copy(m_tmp,aF);
     }
   //+------------------------------------------------------------------+
   //|       .                      |
   //|   a -   (),                       |
   //|   aY -  ,                                         |
   //|   aX -                                          |
   //+------------------------------------------------------------------+
   bool SystemKramer(double  &aK[],double  &aY[],double  &aX[])
     {
      double            m_def=Def(aK);
      if(m_def==0)return(false);
      int               m_KRows,m_KCols;
      GetSize(aK,m_KRows,m_KCols);
      int               m_Size=MathMin(m_KRows,m_KCols);
      int               m_YRows=GetRows(aY);
      m_Size=MathMin(m_Size,m_YRows);
      SetSize(aX,m_Size,1);
      double            m_tmp[];
      for(int m_c=0;m_c<m_Size;m_c++)
        {
         CopyCol(aY,aK,0,m_c,m_tmp);
         aX[m_c]=Def(m_tmp)/m_def;
        }
      return(true);
     }
   //+------------------------------------------------------------------+
   //|        .               |
   //|   a -   (),                       |
   //|   aY -  ,                                         |
   //|   aX -                                          |
   //+------------------------------------------------------------------+
   bool SystemInverse(double  &aK[],double  &aY[],double  &aX[])
     {
      double            m_B[];
      if(!Invert(aK,m_B))return(false);
      MultMx(m_B,aY,aX);
      return(true);
     }
   //+------------------------------------------------------------------+
   //|       .                       |
   //|   a -   (),                       |
   //|   aY -  ,                                         |
   //|   aX -                                          |
   //+------------------------------------------------------------------+
   bool SystemGauss(double  &aK[],double  &aY[],double  &aX[])
     {
      double            m_K[];
      AppendCol(aK,aY,m_K);
      Triangle(m_K,m_K);
      int               m_Rows,m_Cols;
      GetSize(m_K,m_Rows,m_Cols);
      SetSize(aX,m_Rows,1);
      for(int m_r=m_Rows-1;m_r>=0;m_r--)
        {
         double            m_y=GetValue(m_K,m_r,m_Cols-1);
         double            m_k=GetValue(m_K,m_r,m_r);
         if(m_k==0)
           {
            return(false);
           }
         double            m_x=m_y/m_k;
         SetValue(aX,m_r,0,m_x);
         for(int m_r2=m_r-1;m_r2>=0;m_r2--)
           {
            m_k=GetValue(m_K,m_r2,m_r);
            m_y=GetValue(m_K,m_r2,m_Cols-1);
            m_y-=m_k*m_x;
            SetValue(m_K,m_r2,m_Cols-1,m_y);
           }
        }
      return(true);
     }
   //+------------------------------------------------------------------+
   //|     .                                   |
   //|   a -   (),                       |
   //|   aY -  .                                         |
   //|    :                                         |
   //|     -1 -  ,                                            |
   //|      0 -  ,                                           |
   //|      1 -                              |
   //+------------------------------------------------------------------+
   int SystemCheck(double  &aK[],double  &aY[])
     {
      int               m_r0=Rank(aK);
      double            m_K[];
      AppendCol(aK,aY,m_K);
      int               m_r1=Rank(m_K);
      if(m_r0!=m_r1)
        {
         return(-1); //  
        }
      int               m_Rows,m_Cols;
      GetSize(aK,m_Rows,m_Cols);
      int               m_r3=MathMin(m_Rows,m_Cols);
      if(m_r1!=m_r3)
        {
         return(1); //           
        }
      return(0); //        

     }
   //+------------------------------------------------------------------+
   //|        .                         |
   //|      aA       - ,                                         |
   //|      aDigits  -    ,                 |
   //|      aCaption -                                |
   //+------------------------------------------------------------------+
   void Alert(double  &aA[],int aDigits=2,string aCaption="")
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      string            m_str="=== === "+aCaption+" === ===\n";
      for(int m_r=0;m_r<m_ARows;m_r++)
        {
         for(int m_c=0;m_c<m_ACols;m_c++)
           {
            string            m_tstr=DoubleToString(GetValue(aA,m_r,m_c),aDigits);
            m_str=m_str+m_tstr+", ";
           }
         m_str=StringSubstr(m_str,0,StringLen(m_str)-2);
         m_str=m_str+"\n";
        }
      Alert(m_str);
     }
   //+------------------------------------------------------------------+
   //|       ,       |
   //|   , , , ..        |
   //|   :  ,    .      |
   //+------------------------------------------------------------------+
   void Alert2(double  &aA[],int aDigits=2,string aCaption="")
     {
      int               m_ARows,m_ACols;
      GetSize(aA,m_ARows,m_ACols);
      for(int m_r=m_ARows-1;m_r>=0;m_r--)
        {
         string            m_str="";
         for(int m_c=0;m_c<m_ACols;m_c++)
           {
            m_str=m_str+DoubleToString(GetValue(aA,m_r,m_c),aDigits)+", ";
           }
         m_str=StringSubstr(m_str,0,StringLen(m_str)-2);
         Alert(m_str);
        }
      Alert("=== === "+aCaption+" === ===");
     }
   //+------------------------------------------------------------------+
   //|                              |
   //+------------------------------------------------------------------+
   void Alert1Str(double  &aA[],int aDigits=2)
     {
      string            m_str="";
      for(int m_i=0;m_i<ArraySize(aA);m_i++)
        {
         m_str=m_str+DoubleToString(aA[m_i],aDigits)+", ";
        }
      if(m_str!="")
        {
         m_str=StringSubstr(m_str,StringLen(m_str)-2);
        }
      Alert(m_str);
     }

  };
//+------------------------------------------------------------------+